@@ -5,6 +5,7 @@ use App::Ack ();
 use Cwd ();
 use File::Spec ();
 use File::Temp ();
+use Term::ANSIColor ();
 use Test::More;
 
 my $orig_wd;
@@ -22,8 +23,11 @@ sub check_message {
 }
 
 sub prep_environment {
-    delete @ENV{qw( ACK_OPTIONS ACKRC ACK_PAGER HOME )};
-    $orig_wd = Cwd::getcwd();
+    my @ack_args   = qw( ACK_OPTIONS ACKRC ACK_PAGER HOME ACK_COLOR_MATCH ACK_COLOR_FILENAME ACK_COLOR_LINE );
+    my @taint_args = qw( PATH CDPATH IFS );
+    delete @ENV{ @ack_args, @taint_args };
+
+    $orig_wd = getcwd_clean();
 }
 
 sub is_windows {
@@ -34,6 +38,48 @@ sub is_cygwin {
     return $^O eq 'cygwin';
 }
 
+sub is_empty_array {
+    local $Test::Builder::Level = $Test::Builder::Level + 1;
+
+    my $aref = shift;
+    my $msg  = shift;
+
+    my $ok = defined($aref) && (ref($aref) eq 'ARRAY') && (scalar(@{$aref}) == 0);
+
+    if ( !ok( $ok, $msg ) ) {
+        diag( explain( $aref ) );
+    }
+    return $ok;
+}
+
+sub is_nonempty_array {
+    local $Test::Builder::Level = $Test::Builder::Level + 1;
+
+    my $aref = shift;
+    my $msg  = shift;
+
+    my $ok = defined($aref) && (ref($aref) eq 'ARRAY') && (scalar(@{$aref}) > 0);
+
+    if ( !ok( $ok, $msg ) ) {
+        diag( explain( $aref ) );
+    }
+    return $ok;
+}
+
+sub first_line_like {
+    local $Test::Builder::Level = $Test::Builder::Level + 1;
+
+    my $lines = shift;
+    my $re    = shift;
+    my $msg   = shift;
+
+    my $ok = like( $lines->[0], $re, $msg );
+    diag(explain($lines)) unless $ok;
+
+    return $ok;
+}
+
+
 sub build_ack_invocation {
     my @args = @_;
 
@@ -139,8 +185,7 @@ sub run_ack {
         fail( q{Automatically fail stderr check for TODO tests.} );
     }
     else {
-        is( scalar @{$stderr}, 0, "Should have no output to stderr: ack @args" )
-            or diag( join( "\n", "STDERR:", @{$stderr} ) );
+        is_empty_array( $stderr, "Should have no output to stderr: ack @args" );
     }
 
     return wantarray ? @{$stdout} : join( "\n", @{$stdout} );
@@ -174,6 +219,8 @@ sub run_cmd {
 
     record_option_coverage(@cmd);
 
+    check_command_for_taintedness( @cmd );
+
     my ( @stdout, @stderr );
 
     if ( is_windows() ) {
@@ -259,13 +306,13 @@ sub run_cmd {
             close $stdout_read;
             close $stderr_read;
 
-            if(my $input = $options->{'input'}) {
-                # XXX check error
-                open STDIN, '-|', @$input;
+            if (my $input = $options->{input}) {
+                check_command_for_taintedness( @{$input} );
+                open STDIN, '-|', @{$input} or die "Can't open STDIN: $!";
             }
 
-            open STDOUT, '>&', $stdout_write;
-            open STDERR, '>&', $stderr_write;
+            open STDOUT, '>&', $stdout_write or die "Can't open STDOUT: $!";
+            open STDERR, '>&', $stderr_write or die "Can't open STDERR: $!";
 
             exec @cmd;
         }
@@ -294,12 +341,14 @@ sub run_ack_with_stderr {
     my @stdout;
     my @stderr;
 
+    my $perl = caret_X();
+
     @args = build_ack_invocation( @args );
     if ( $ENV{'ACK_TEST_STANDALONE'} ) {
-        unshift( @args, $^X );
+        unshift( @args, $perl );
     }
     else {
-        unshift( @args, $^X, "-Mblib=$orig_wd" );
+        unshift( @args, $perl, "-Mblib=$orig_wd" );
     }
 
     return run_cmd( @args );
@@ -327,7 +376,7 @@ sub pipe_into_ack_with_stderr {
         # to simply using 'cat'.  The answer is that we have no
         # idea what the user's environment is like; they may not
         # have cat.  We *do* know, however, that they have perl.
-        input => [$^X, '-pe1', $input],
+        input => [caret_X(), '-pe1', $input],
     });
 }
 
@@ -351,7 +400,7 @@ sub lists_match {
         $path = File::Next::reslash( $path ); ## no critic (Variables::ProhibitPackageVars)
     }
 
-    return subtest "lists_match($msg)" => sub {
+    return subtest "lists_match( $msg )" => sub {
         plan tests => 1;
 
         my $rc = eval 'use Test::Differences; 1;';
@@ -381,7 +430,7 @@ sub ack_lists_match {
 
     my @results = run_ack( @args );
 
-    return subtest "ack_lists_match($message)" => sub {
+    return subtest "ack_lists_match( $message )" => sub {
         plan tests => 1;
 
         my $ok = lists_match( \@results, $expected, $message );
@@ -397,7 +446,7 @@ sub sets_match {
     my @expected = @{+shift};
     my $msg      = check_message( shift );
 
-    return subtest "sets_match($msg)" => sub {
+    return subtest "sets_match( $msg )" => sub {
         plan tests => 1;
 
         return lists_match( [sort @actual], [sort @expected], $msg );
@@ -432,13 +481,14 @@ sub record_option_coverage {
 
     my $record_options = File::Spec->catfile($orig_wd, 'record-options');
 
+    my $perl = caret_X();
     if ( @command_line == 1 ) {
         my $command_line = $command_line[0];
 
         # strip the command line up until 'ack' is found
         $command_line =~ s/^.*ack\b//;
 
-        $command_line = "$^X $record_options $command_line";
+        $command_line = "$perl -T $record_options $command_line";
 
         system $command_line;
     }
@@ -447,7 +497,7 @@ sub record_option_coverage {
             shift @command_line;
         }
         shift @command_line; # get rid of 'ack' itself
-        unshift @command_line, $^X, $record_options;
+        unshift @command_line, $perl, '-T', $record_options;
 
         system @command_line;
     }
@@ -455,6 +505,39 @@ sub record_option_coverage {
     return;
 }
 
+=head2 colorize( $big_long_string )
+
+Turns a multi-line input string into its corresponding array of lines, with colorized transformations.
+
+    <text> gets turned to filename color.
+    {text} gets turned to line number color.
+    (text) gets turned to highlight color.
+
+=cut
+
+sub colorize {
+    my $input = shift;
+
+    my @lines = split( /\n/, $input );
+
+    for my $line ( @lines ) {
+        # File name
+        $line =~ s/<(.+?)>/Term::ANSIColor::colored($1, 'bold green')/eg;
+
+        # Line number
+        $line =~ s/\{(.+?)\}/Term::ANSIColor::colored($1, 'bold yellow')/eg;
+
+        # Matches
+        my $n;
+        $n += $line =~ s/\((.+?)\)/Term::ANSIColor::colored($1, 'black on_yellow')/eg;
+
+        $line .= "\033[0m\033[K" if $n;
+    }
+
+    return @lines;
+}
+
+
 BEGIN {
     my $has_io_pty = eval {
         require IO::Pty;
@@ -515,12 +598,21 @@ BEGIN {
                 }
                 $slave->set_raw();
 
-                open STDIN,  '<&', $slave->fileno();
-                open STDOUT, '>&', $slave->fileno();
-                open STDERR, '>&', $slave->fileno();
+                open STDIN,  '<&', $slave->fileno() or die "Can't open: $!";
+                open STDOUT, '>&', $slave->fileno() or die "Can't open: $!";
+                open STDERR, '>&', $slave->fileno() or die "Can't open: $!";
 
                 close $slave;
 
+                my $perl = caret_X();
+
+                if ( $ENV{'ACK_TEST_STANDALONE'} ) {
+                    unshift( @cmd, $perl );
+                }
+                else {
+                    unshift( @cmd, $perl, "-Mblib=$orig_wd" );
+                }
+
                 exec @cmd;
             }
         };
@@ -641,4 +733,57 @@ sub get_options {
     );
 }
 
+
+# This is just a handy diagnostic tool.
+sub check_command_for_taintedness {
+    my @args = @_;
+
+    my $bad = 0;
+
+    my @tainted;
+    for my $arg ( @args ) {
+        if ( is_tainted( $arg ) ) {
+            push( @tainted, $arg );
+        }
+    }
+
+    if ( @tainted ) {
+        die "Can't execute this command because of taintedness:\nAll args: @args\nTainted:  @tainted\n";
+    }
+
+    return;
+}
+
+
+sub is_tainted {
+    no warnings qw(void uninitialized);
+
+    return !eval { local $SIG{__DIE__} = 'DEFAULT'; join('', shift), kill 0; 1 };
+}
+
+
+sub untaint {
+    my ( $s ) = @_;
+
+    $s =~ /\A(.*)\z/;
+    return $1;
+}
+
+
+sub caret_X {
+    # XXX How is it $^X can be tainted?  We should not have to untaint it.
+    $^X =~ /(.+)/;
+    my $perl = $1;
+
+    return $perl;
+}
+
+
+sub getcwd_clean {
+    # XXX How is it that this guy is tainted?
+    my $wd = Cwd::getcwd();
+    $wd =~ /(.+)/;
+    return $1;
+}
+
 1;
